# Copyright (c) 2012 Leif Johnson <leif@leifjohnson.net>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

'''This file contains optimization methods for neural networks.'''

import climate
import collections
import itertools
import numpy as np
import numpy.random as rng
import theano
import theano.tensor as TT
import sys

from . import dataset
from . import feedforward
from . import recurrent

logging = climate.get_logger(__name__)


class Trainer(object):
    '''This is a base class for all trainers.'''

    def update_params(self, targets):
        for param, target in zip(self.params, targets):
            param.set_value(target)


class SGD(Trainer):
    '''Stochastic gradient descent network trainer.'''

    def __init__(self, network, **kwargs):
        self.params = network.params(**kwargs)
        self.validation_frequency = kwargs.get('validate', 3)
        self.min_improvement = kwargs.get('min_improvement', 0.)
        self.iterations = kwargs.get('num_updates', 1000)
        self.patience = kwargs.get('patience', 100)
        self.momentum = kwargs.get('momentum', 0.)
        self.grad_clip = kwargs.get('gradient_clip', 1e5)
        self.learning_rate = kwargs.get('learning_rate', 0.1)
        self.learning_rate_decay = kwargs.get('learning_rate_decay', 0.1)

        J = network.J(**kwargs)
        costs = [J]
        self.cost_names = ['J']
        for name, monitor in network.monitors:
            self.cost_names.append(name)
            costs.append(monitor)
        self.f_train = theano.function(network.inputs, costs, updates=network.updates)
        self.f_grad = theano.function(network.inputs, [TT.grad(J, p) for p in self.params])
        self.f_eval = theano.function(network.inputs, costs)

    def train(self, train_set, valid_set=None, **kwargs):
        '''We train over mini-batches and evaluate periodically.'''
        best_cost = 1e100
        best_iter = 0
        best_params = []
        velocities = []
        P = len(self.params)
        for p in self.params:
            v = p.get_value()
            best_params.append(v.copy())
            velocities.append(np.zeros_like(v))

        learn = self._nag if self.momentum > 0 else self._sgd

        for i in xrange(self.iterations):
            # if it's time, evaluate the model on the validation dataset.
            if not i % self.validation_frequency:
                try:
                    costs = np.mean([self.f_eval(*x) for x in valid_set], axis=0)
                except KeyboardInterrupt:
                    logging.info('interrupted !')
                    break
                marker = ''
                if (best_cost - costs[0]) / best_cost > self.min_improvement:
                    best_cost = costs[0]
                    best_iter = i
                    best_params = [p.get_value().copy() for p in self.params]
                    marker = ' *'
                else:
                    self.learning_rate *= 1 - self.learning_rate_decay

                cost_desc = ' '.join(
                    '%s=%.4f' % i for i in zip(self.cost_names, costs))
                logging.info('SGD %i -- valid %s%s', i + 1, cost_desc, marker)

                if i - best_iter > self.patience:
                    logging.error('patience elapsed, bailing out')
                    break

            costs = []
            grads = []
            try:
                for costs_, grads_ in learn(train_set, velocities):
                    costs.append(costs_)
                    grads.append(grads_)
            except KeyboardInterrupt:
                logging.info('interrupted !')
                break

            cost_desc = ' '.join(
                '%s=%.4f' % i for i in
                zip(self.cost_names, np.mean(costs, axis=0)))
            grad_desc = ' '.join(
                '%s=%.4f' % (p.name, x) for p, x in
                zip(self.params, np.mean(grads, axis=0)))
            logging.info('SGD %i/%i @%.2e -- train %s -- grad %s',
                         i + 1, self.iterations, self.learning_rate,
                         cost_desc, grad_desc)

        self.update_params(best_params)

    def _nag(self, train_set, velocities):
        '''Make one run through the training set.

        We update parameters after each minibatch according to Nesterov's
        Accelerated Gradient. The basic difference between NAG and "classical"
        momentum is that NAG computes the gradients at the position in parameter
        space where "classical" momentum would put us at the next step.

        In theory, this helps correct for oversteps during learning. If momentum
        would lead us to overshoot, then the gradient at that overshot place
        will point backwards, toward where we came from.
        '''
        gc = self.grad_clip
        # TODO: run this loop in parallel !
        for x in train_set:
            grads = []
            moves = []
            # first, move to the position in parameter space that we would get
            # to using classical momentum-based sgd.
            for param, vel in zip(self.params, velocities):
                u = self.momentum * vel
                v = param.get_value(borrow=True)
                v += u
                param.set_value(v, borrow=True)
                moves.append(u)
            for param, vel, grad, u in zip(self.params, velocities, self.f_grad(*x), moves):
                # measure the gradient at this new position.
                g = np.asarray(grad)
                grads.append(np.linalg.norm(g))
                # update the velocity using the new gradient. remember that
                # u = self.momentum * vel.
                np.clip(u - self.learning_rate * g, -gc, gc, out=vel)
                # subtract out the movement from momentum that we added in
                # above, and add the updated velocity.
                v = param.get_value(borrow=True)
                v += vel - u
                param.set_value(v, borrow=True)
            yield self.f_train(*x), grads

    def _sgd(self, train_set, velocities):
        '''Make one run through the training set.

        We update parameters after each minibatch according to standard
        stochastic gradient descent.
        '''
        # TODO: run this loop in parallel !
        for x in train_set:
            grads = []
            for param, grad in zip(self.params, self.f_grad(*x)):
                g = np.asarray(grad)
                grads.append(np.linalg.norm(g))
                v = param.get_value(borrow=True)
                v -= self.learning_rate * g
                param.set_value(v, borrow=True)
            yield self.f_train(*x), grads


class CG(Trainer):
    '''Conjugate gradient trainer for neural networks.'''

    def __init__(self, network, **kwargs):
        raise NotImplementedError


class LM(Trainer):
    '''Levenberg-Marquardt trainer for neural networks.

    Based on the description of the algorithm in "Levenberg-Marquardt
    Optimization" by Sam Roweis.
    '''

    def __init__(self, network, **kwargs):
        raise NotImplementedError


class HF(Trainer):
    '''The hessian free trainer shells out to an external implementation.

    hf.py was implemented by Nicholas Boulanger-Lewandowski and made available
    to the public (yay !). If you don't have a copy of the module handy, this
    class will attempt to download it from github.
    '''

    URL = 'https://raw.github.com/boulanni/theano-hf/master/hf.py'

    def __init__(self, network, **kwargs):
        import os, tempfile, urllib
        sys.path.append(tempfile.gettempdir())

        try:
            import hf
        except:
            # if hf failed to import, try downloading it and saving it locally.
            logging.error('hf import failed, attempting to download %s', HF.URL)
            path = os.path.join(tempfile.gettempdir(), 'hf.py')
            urllib.urlretrieve(HF.URL, path)
            logging.error('downloaded hf code to %s', path)
            import hf

        self.params = network.params(**kwargs)
        self.opt = hf.hf_optimizer(
            self.params,
            network.inputs,
            network.y,
            [network.J(**kwargs)] + [mon for _, mon in network.monitors],
            network.hiddens[-1] if isinstance(network, recurrent.Network) else None)

        # fix mapping from kwargs into a dict to send to the hf optimizer
        kwargs['validation_frequency'] = kwargs.pop('validate', sys.maxint)
        for k in set(kwargs) - set(self.opt.train.im_func.func_code.co_varnames[1:]):
            kwargs.pop(k)
        self.kwargs = kwargs

    def train(self, train_set, valid_set=None, **kwargs):
        self.update_params(self.opt.train(
            train_set, kwargs['cg_set'], validation=valid_set, **self.kwargs))


class Sample(Trainer):
    '''This trainer replaces network weights with samples from the input.'''

    @staticmethod
    def reservoir(xs, n):
        '''Select a random sample of n items from xs.'''
        pool = []
        for i, x in enumerate(xs):
            if len(pool) < n:
                pool.append(x / np.linalg.norm(x))
                continue
            j = rng.randint(i + 1)
            if j < n:
                pool[j] = x / np.linalg.norm(x)
        # if the pool still has fewer than n items, pad with distorted random
        # duplicates from the source data.
        L = len(pool)
        while len(pool) < n:
            x = pool[rng.randint(L)]
            pool.append(x + np.std(pool, axis=0) * rng.randn(*x.shape))
        return pool

    def __init__(self, network, **kwargs):
        self.network = network

    def train(self, train_set, valid_set=None, **kwargs):
        ifci = itertools.chain.from_iterable

        # set output (decoding) weights on the network.
        last = lambda x: x[-1] if isinstance(x, (tuple, list)) else x
        samples = ifci(last(t) for t in train_set)
        for w in self.network.weights:
            k, n = w.get_value(borrow=True).shape
            if w.name.startswith('W_out_'):
                arr = np.vstack(Sample.reservoir(samples, k))
                logging.info('setting weights for %s: %d x %d <- %s', w.name, k, n, arr.shape)
                w.set_value(arr)

        # set input (encoding) weights on the network.
        first = lambda x: x[0] if isinstance(x, (tuple, list)) else x
        samples = ifci(first(t) for t in train_set)
        for i, h in enumerate(self.network.hiddens):
            w = self.network.weights[i]
            m, k = w.get_value(borrow=True).shape
            arr = np.vstack(Sample.reservoir(samples, k)).T
            logging.info('setting weights for %s: %d x %d <- %s', w.name, m, k, arr.shape)
            w.set_value(arr)
            samples = ifci(self.network.feed_forward(first(t))[i-1] for t in train_set)


class Layerwise(Trainer):
    '''This trainer adapts parameters using a variant of layerwise pretraining.

    In this variant, we create "taps" at increasing depths into the original
    network weights, training only those weights that are below the tap. So, for
    a hypothetical binary classifier network with layers [3, 4, 5, 6, 2], we
    would first insert a tap after the first hidden layer (effectively a binary
    classifier in a [3, 4, 2] configuration) and train just that network. Then
    we insert a tap at the next layer (effectively training a [3, 4, 5, 2]
    classifier), and so forth.

    By inserting taps into the original network, we preserve all of the relevant
    settings of noise, dropouts, loss function and the like, in addition to
    obviating the need for copying trained weights around between different
    Network instances.
    '''

    def __init__(self, network, **kwargs):
        self.network = network
        self.kwargs = kwargs

    def train(self, train_set, valid_set=None, **kwargs):
        y = self.network.y
        hiddens = list(self.network.hiddens)
        weights = list(self.network.weights)
        biases = list(self.network.biases)

        nout = len(biases[-1].get_value(borrow=True))
        nhids = [len(b.get_value(borrow=True)) for b in biases]
        for i in range(1, len(nhids)):
            W, b, _ = self.network._weights_and_bias(nhids[i-1], nout, 'lwout-%d' % i)
            self.network.y = TT.dot(hiddens[i-1], W) + b
            self.network.hiddens = hiddens[:i]
            self.network.weights = weights[:i] + [W]
            self.network.biases = biases[:i] + [b]
            SGD(self.network, **self.kwargs).train(train_set, valid_set)
            self.network.save('/tmp/layerwise-%s-h%f-n%f-d%f-w%f-%d.pkl.gz' % (
                    ','.join(map(str, self.kwargs['layers'])),
                    self.kwargs['hidden_l1'],
                    self.kwargs['input_noise'],
                    self.kwargs['hidden_dropouts'],
                    self.kwargs['weight_l1'],
                    i))

        self.network.y = y
        self.network.hiddens = hiddens
        self.network.weights = weights
        self.network.biases = biases


class FORCE(Trainer):
    '''FORCE is a training method for recurrent nets by Sussillo & Abbott.

    This implementation needs some more love before it will work.
    '''

    def __init__(self, network, **kwargs):
        W_in, W_pool, W_out = network.weights

        n = W_pool.get_value(borrow=True).shape[0]
        self.alpha = kwargs.get('learning_rate', 1. / n)
        P = theano.shared(np.eye(n).astype(FLOAT) * self.alpha)

        k = TT.dot(P, network.state)
        rPr = TT.dot(network.state, k)
        c = 1. / (1. + rPr)
        dw = network.error(**kwargs) * c * k

        J = network.J(**kwargs)
        updates = {}
        updates[P] = P - c * TT.outer(k, k)
        updates[W_pool] = W_pool - dw
        updates[W_out] = W_out - dw
        updates[b_out] = b_out - self.alpha * TT.grad(J, b_out)

        costs = [J] + network.monitors
        self.f_eval = theano.function(network.inputs, costs)
        self.f_train = theano.function(network.inputs, costs, updates=updates)

    @property
    def learning_rate(self):
        return self.alpha



# Copyright (c) 2012 Leif Johnson <leif@leifjohnson.net>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

'''This file contains recurrent network structures.'''

import climate
import numpy as np
import numpy.random as rng
import theano
import theano.tensor as TT

from theano.sandbox.rng_mrg import MRG_RandomStreams as RandomStreams

from . import feedforward as ff

logging = climate.get_logger(__name__)


class Network(ff.Network):
    '''A fully connected recurrent network with one input and one output layer.
    '''

    def __init__(self, layers, activation, rng=None, input_noise=0,
                 hidden_noise=0, input_dropouts=0, hidden_dropouts=0, **kwargs):
        '''Initialize the weights and computation graph for a recurrent network.

        layers: A sequence of three integers specifying the number of units in
          the input, hidden, and output layers, respectively.
        activation: A callable that takes one argument (a matrix) and returns
          another matrix. This is the activation function that each hidden unit
          in the network uses.
        rng: Use a specific Theano random number generator. A new one will be
          created if this is None.
        input_noise: Standard deviation of desired noise to inject into input.
        hidden_noise: Standard deviation of desired noise to inject into
          hidden unit activation output.
        input_dropouts: Proportion of input units to randomly set to 0.
        hidden_dropouts: Proportion of hidden unit activations to randomly set
          to 0.

        Recognized keyword arguments:

        pool_damping: This parameter (a float in [0, 1]) governs the proportion
          of past state information retained by the recurrent pool neurons.
        pool_noise: Add gaussian noise to recurrent pool neurons with this
          variance.
        pool_dropouts: Randomly set the state of this fraction (a float in
          [0, 1]) of recurrent pool neurons to zero.
        pool_error_start: Compute error metrics starting at this time step.
        '''
        self.rng = rng or RandomStreams()
        pool_noise = kwargs.get('pool_noise', 0.)
        pool_dropouts = kwargs.get('pool_dropouts', 0.)
        pool_damping = kwargs.get('pool_damping', 0.)
        self.error_start = kwargs.get('pool_error_start', 3)

        # in this module, x refers to a network's input, and y to its output.
        self.x = TT.matrix('x')

        parameter_count = 0
        layers = list(layers)
        num_out = layers.pop()
        num_pool = layers.pop()

        W_in = []
        b_in = []
        for i, (a, b) in enumerate(zip(layers, layers[1:] + [num_pool])):
            W, b, params = self._weights_and_bias(a, b, i)
            parameter_count += params
            W_in.append(W)
            b_in.append(b)

        # discard bias for last input layer (we create it explicitly below, as
        # it overlaps with the bias for the recurrent pool)
        b_in.pop()
        parameter_count -= num_pool

        W_pool, b_pool, params = self._weights_and_bias(num_pool, num_pool, 'pool')
        parameter_count += params

        W_out, b_out, params = self._weights_and_bias(num_pool, num_out, 'out')
        parameter_count += params

        logging.info('%d total network parameters', parameter_count)

        def step(x_t, h_tm1):
            z = self._noise_and_dropout(x_t, input_noise, input_dropouts)
            encs = []
            for W, b in zip(W_in, b_in):
                encs.append(self._noise_and_dropout(
                    activation(TT.dot(z, W) + b), hidden_noise, hidden_dropouts))
                z = encs[-1]
            h = activation(TT.dot(z, W_in[-1]) + TT.dot(h_tm1, W_pool) + b_pool)
            h_t = self._noise_and_dropout(
                (1 - pool_damping) * h + pool_damping * h_tm1,
                pool_noise, pool_dropouts)
            return encs + [h_t, TT.dot(h_t, W_out) + b_out]

        h_0 = TT.zeros((num_pool, ), dtype=ff.FLOAT)
        outputs, self.updates = theano.scan(
            fn=step, sequences=self.x,
            outputs_info=[None for _ in b_in] + [h_0, None])

        self.y = outputs.pop()
        self.hiddens = outputs
        self.weights = W_in + [W_pool, W_out]
        self.biases = b_in + [b_pool, b_out]

        # compute a complete pass over an input sequence.
        self.feed_forward = theano.function(
            [self.x], self.hiddens + [self.y], updates=self.updates)


class Autoencoder(Network):
    '''An autoencoder attempts to reproduce its input.'''

    @property
    def cost(self):
        err = self.y - self.x
        return TT.mean((err * err).sum(axis=1)[self.error_start:])


class Regressor(Network):
    '''A regressor attempts to produce a target output.'''

    def __init__(self, *args, **kwargs):
        super(Regressor, self).__init__(*args, **kwargs)
        self.k = TT.matrix('k')

    @property
    def inputs(self):
        return [self.x, self.k]

    @property
    def cost(self):
        err = self.y - self.k
        return TT.mean((err * err).sum(axis=1)[self.error_start:])



# Copyright (c) 2012 Leif Johnson <leif@leifjohnson.net>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

'''This module contains a number of classes for modeling neural nets in Theano.
'''

import climate
import pickle
import gzip
import numpy as np
import theano
import theano.tensor as TT

from theano.sandbox.rng_mrg import MRG_RandomStreams as RandomStreams

logging = climate.get_logger(__name__)

FLOAT = theano.config.floatX
DEBUG_BATCH_SIZE = 11


class Network(object):
    '''The network class is a fully-connected feedforward net.

    It can add random noise to the inputs, or to the hidden unit activations.
    Adding noise to these quantities can be seen as a form of regularization.

    This class also permits "decoding" (computing final network output) from
    more than just the final hidden layer of units ; however, decoding must
    always include the final k hidden layers in the network.

    Attributes
    ----------
    hiddens : list of floats
    weights : list of Theano variables
    biases : list of Theano variables
    '''

    def __init__(self, layers, activation, rng=None, input_noise=0,
                 hidden_noise=0, input_dropouts=0, hidden_dropouts=0,
                 **kwargs):
        '''Create a new feedforward network of a specific topology.

        Arguments
        ---------
        layers : sequence of int
          A sequence of integers specifying the number of units at each layer.
          As an example, layers=(10, 20, 3) has one "input" layer with 10 units,
          one "hidden" layer with 20 units, and one "output" layer with 3 units.
          That is, inputs should be of length 10, and outputs will be of length
          3.
        activation : callable(numeric) -> numeric
          A callable that takes an array of values (specifically, an array of
          "raw unit inputs") and returns another array of values of the same
          dimension (specifically, an array of "unit activations"). This is the
          activation function that each hidden unit in the network uses.
        rng : theano RandomStreams object, optional
          Use a specific Theano random number generator. A new one will be
          created if this is None.
        input_noise : float
          Standard deviation of desired noise to inject into input.
        hidden_noise : float
          Standard deviation of desired noise to inject into hidden unit
          activation output.
        input_dropouts : float in [0, 1]
          Proportion of input units to randomly set to 0.
        hidden_dropouts : float in [0, 1]
          Proportion of hidden unit activations to randomly set to 0.

        Keyword arguments
        -----------------
        decode : positive int
          Any of the hidden layers can be tapped at the output. Just specify a
          value greater than 1 to tap the last N hidden layers.
        tied_weights : bool
          Construct decoding weights using the transpose of the encoding weights
          on corresponding layers. If not True, decoding weights will be
          constructed using a separate weight matrix.
        '''
        self.hiddens = []
        self.weights = []
        self.biases = []

        self.rng = rng or RandomStreams()
        tied_weights = kwargs.get('tied_weights')
        decode = kwargs.get('decode', 1)

        # in this module, x refers to a network's input, and y to its output.
        self.x = TT.matrix('x')
        self.x.tag.test_value = np.random.randn(DEBUG_BATCH_SIZE, layers[0])

        parameter_count = 0
        sizes = layers[:-1]

        if tied_weights:
            error = 'with --tied-weights, --layers must be a palindrome of length 2k+1'
            assert len(layers) % 2 == 1, error
            k = len(layers) // 2
            encode = np.asarray(layers[:k])
            decode = np.asarray(layers[k+1:])
            assert np.allclose(encode - decode[::-1], 0), error
            sizes = layers[:k+1]

        z = self._noise_and_dropout(self.x, input_noise, input_dropouts)
        for i, (a, b) in enumerate(zip(sizes[:-1], sizes[1:])):
            Wi, bi, params = self._weights_and_bias(a, b, i)
            parameter_count += params
            self.hiddens.append(self._noise_and_dropout(
                activation(TT.dot(z, Wi) + bi), hidden_noise, hidden_dropouts))
            self.weights.append(Wi)
            self.biases.append(bi)
            z = self.hiddens[-1]
            z.tag.test_value = np.random.randn(DEBUG_BATCH_SIZE, b)

        w = len(self.weights)
        if tied_weights:
            for i in range(w - 1, -1, -1):
                h = self.hiddens[-1]
                a, b = self.weights[i].get_value(borrow=True).shape
                logging.info('tied weights from layer %d: %s x %s', i, b, a)
                self.hiddens.append(TT.dot(h - TT.neq(h, 0) * self.biases[i], self.weights[i].T))
        else:
            n = layers[-1]
            decoders = []
            for i in range(w - 1, w - 1 - decode, -1):
                b = self.biases[i].get_value(borrow=True).shape[0]
                Di, _, params = self._weights_and_bias(b, n, 'out_%d' % i)
                parameter_count += params - n
                decoders.append(TT.dot(self.hiddens[i], Di))
                self.weights.append(Di)
            parameter_count += n
            bias = theano.shared(np.zeros((n, ), FLOAT), name='bias_out')
            bias.tag.test_value = np.random.randn(n)
            self.biases.append(bias)
            self.hiddens.append(sum(decoders) + bias)

        logging.info('%d total network parameters', parameter_count)

        self.y = self.hiddens.pop()

        self.updates = {}

        # calling this computes a forward pass, returning all layer activations.
        self.feed_forward = theano.function([self.x], self.hiddens + [self.y])

    @property
    def inputs(self):
        '''Return a list of Theano input variables for this network.'''
        return [self.x]

    @property
    def monitors(self):
        '''Generate a sequence of name-value pairs for monitoring the network.
        '''
        yield 'error', self.cost
        for i, s in enumerate(self.sparsities):
            yield 'sparse_{}'.format(i+1), s

    @property
    def sparsities(self):
        '''Return the sparsity (% of 0-activations) in each hidden layer.'''
        return [TT.eq(h, 0).mean() for h in self.hiddens]

    @staticmethod
    def _weights_and_bias(a, b, suffix):
        '''Create a layer of weights and bias values.

        Arguments
        ---------
        a : int
          Number of rows of the weight matrix -- also, the number of "input"
          units that the weight matrix connects.
        b : int
          Number of columns of the weight matrix -- also, the number of "output"
          units that the weight matrix connects.
        suffix : str
          A string suffix to use in the Theano name for the created variables.
          This string will be appended to "W_" (for the weights) and "b_" (for
          the biases) parameters that are created and returned.

        Returns
        -------
        weight : Theano shared array
          A shared array containing Theano values representing the weights
          connecting each "input" unit to each "output" unit.
        bias : Theano shared array
          A shared array containing Theano values representing the bias values
          on each of the "output" units.
        count : int
          The number of parameters that are included in the returned variables.
        '''
        arr = np.random.randn(a, b) / np.sqrt(a + b)
        weight = theano.shared(arr.astype(FLOAT), name='W_%s' % suffix)
        weight.tag.test_value = np.random.randn(a, b)
        bias = theano.shared(np.zeros((b, ), FLOAT), name='b_%s' % suffix)
        bias.tag.test_value = np.random.randn(b)
        logging.info('weights for layer %s: %s x %s', suffix, a, b)
        return weight, bias, (a + 1) * b

    def _noise_and_dropout(self, x, sigma, rho):
        '''Add noise and dropouts to elements of x as needed.

        Arguments
        ---------
        x : Theano array
          Input array to add noise and dropouts to.
        sigma : float
          Standard deviation of noise to add to x. If this is 0, then no noise
          is added to the values of x.
        rho : float, in [0, 1]
          Fraction of elements of x to set randomly to 0. If this is 0, then no
          elements of x are set randomly to 0.

        Returns
        -------
        The parameter x, plus additional noise as specified.
        '''
        if sigma > 0 and rho > 0:
            noise = self.rng.normal(size=x.shape, std=sigma)
            mask = self.rng.binomial(size=x.shape, n=1, p=1-rho, dtype=FLOAT)
            return mask * (x + noise)
        if sigma > 0:
            return x + self.rng.normal(size=x.shape, std=sigma)
        if rho > 0:
            mask = self.rng.binomial(size=x.shape, n=1, p=1-rho, dtype=FLOAT)
            return mask * x
        return x

    def params(self, **kwargs):
        '''Return a list of the Theano parameters for this network.'''
        params = []
        params.extend(self.weights)
        if kwargs.get('no_learn_biases'):
            pass
        else:
            params.extend(self.biases)
        return params

    def predict(self, x):
        '''Compute a forward pass of the inputs, returning the net output.

        Returns
        -------
        Mathematically, a neural network is a specific parametrization of a
        generic function approximator. The network ultimately attempts to
        compute some function value g at the given input x: g(x). Due to noise
        in the training process, however, the value computed by the network
        could differ from the "true" value of g(x).
        '''
        return self.feed_forward(x)[-1]

    __call__ = predict

    def save(self, filename):
        '''Save the parameters of this network to disk.

        Arguments
        ---------
        filename : str
          Save the parameters of this network to a pickle file at the named
          path. If this name ends in ".gz" then the output will automatically be
          gzipped; otherwise the output will be a "raw" pickle.
        '''
        opener = gzip.open if filename.lower().endswith('.gz') else open
        handle = opener(filename, 'wb')
        pickle.dump(
            dict(weights=[p.get_value().copy() for p in self.weights],
                 biases=[p.get_value().copy() for p in self.biases],
                 ), handle, -1)
        handle.close()
        logging.info('%s: saved model parameters', filename)

    def load(self, filename):
        '''Load the parameters for this network from disk.

        Arguments
        ---------
        filename : str
          Load the parameters of this network from a pickle file at the named
          path. If this name ends in ".gz" then the input will automatically be
          gunzipped; otherwise the input will be treated as a "raw" pickle.
        '''
        opener = gzip.open if filename.lower().endswith('.gz') else open
        handle = opener(filename, 'rb')
        params = pickle.load(handle)
        for target, source in zip(self.weights, params['weights']):
            target.set_value(source)
        for target, source in zip(self.biases, params['biases']):
            target.set_value(source)
        handle.close()
        logging.info('%s: loaded model parameters', filename)

    def J(self, weight_l1=0, weight_l2=0, hidden_l1=0, hidden_l2=0, **unused):
        '''Return a cost function for this network.

        Arguments
        ---------
        weight_l1 : float
          Regularize the L1 norm of unit connection weights by this constant.
        weight_l2 : float
          Regularize the L2 norm of unit connection weights by this constant.
        hidden_l1 : float
          Regularize the L1 norm of hidden unit activations by this constant.
        hidden_l2 : float
          Regularize the L2 norm of hidden unit activations by this constant.

        Returns
        -------
        A Theano symbol that can be incorporated into a Theano function.
        '''
        cost = self.cost
        if weight_l1 > 0:
            cost += weight_l1 * sum(abs(w).sum() for w in self.weights)
        if weight_l2 > 0:
            cost += weight_l2 * sum((w * w).sum() for w in self.weights)
        if hidden_l1 > 0:
            cost += hidden_l1 * sum(abs(h).mean(axis=0).sum() for h in self.hiddens)
        if hidden_l2 > 0:
            cost += hidden_l2 * sum((h * h).mean(axis=0).sum() for h in self.hiddens)
        return cost


class Autoencoder(Network):
    '''An autoencoder attempts to reproduce its input.'''

    @property
    def cost(self):
        err = self.y - self.x
        return TT.mean((err * err).sum(axis=1))


class Regressor(Network):
    '''A regressor attempts to produce a target output.'''

    def __init__(self, *args, **kwargs):
        self.k = TT.matrix('k')
        super(Regressor, self).__init__(*args, **kwargs)
        # for shape debugging
        w = self.weights[len(self.biases) - 1]
        self.k.tag.test_value = np.random.randn(
            DEBUG_BATCH_SIZE, w.get_value(borrow=True).shape[1])

    @property
    def inputs(self):
        return [self.x, self.k]

    @property
    def cost(self):
        err = self.y - self.k
        return TT.mean((err * err).sum(axis=1))


class Classifier(Network):
    '''A classifier attempts to match a 1-hot target output.'''

    def __init__(self, *args, **kwargs):
        self.k = TT.ivector('k')
        self.k.tag.test_value = (3 * np.random.randn(DEBUG_BATCH_SIZE)).astype('int32')

        super(Classifier, self).__init__(*args, **kwargs)

        self.y = self.softmax(self.y)

        # recompile feed_forward method to return proper y value.
        self.feed_forward = theano.function([self.x], self.hiddens + [self.y])

    @staticmethod
    def softmax(x):
        # TT.nnet.softmax doesn't work with the HF trainer.
        z = TT.exp(x.T - x.T.max(axis=0))
        return (z / z.sum(axis=0)).T

    @property
    def inputs(self):
        return [self.x, self.k]

    @property
    def cost(self):
        return -TT.mean(TT.log(self.y)[TT.arange(self.k.shape[0]), self.k])

    @property
    def incorrect(self):
        return TT.mean(TT.neq(TT.argmax(self.y, axis=1), self.k))

    @property
    def monitors(self):
        yield 'incorrect', self.incorrect
        for i, s in enumerate(self.sparsities):
            yield 'sparse_{}'.format(i+1), s

    def classify(self, x):
        return self.predict(x).argmax(axis=1)



# Copyright (c) 2012 Leif Johnson <leif@leifjohnson.net>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

'''This file contains a class for handling batched datasets.'''

import climate
import numpy.random as rng

logging = climate.get_logger(__name__)


class SequenceDataset(object):
    '''This class handles batching and shuffling a dataset.

    It's mostly copied from the dataset class from hf.py, except that the
    constructor has slightly different semantics.
    '''

    def __init__(self, *data, **kwargs):
        '''Create a minibatch dataset from a number of different data arrays.

        Positional arguments:

        There should be one unnamed keyword argument for each input in the
        neural network that will be processing this dataset. For instance, if
        you are dealing with a classifier network, you'll need one argument for
        the inputs (e.g., mnist digit pixels), and another argument for the
        target outputs (e.g., digit class labels). The order of the arguments
        should be the same as the order of inputs in the network. All arguments
        are expected to have the same number of elements along the first axis.

        Alternatively, if there is only one positional arg, and it is callable,
        then that callable will be invoked repeatedly at training and test time.
        Each invocation of the callable should return a tuple containing one
        minibatch of data. The callable will not be passed any arguments.

        Keyword arguments:

        size or batch_size: The size of the mini-batches to create from the
          data sequences. Defaults to 32.
        batches: The number of batches to yield for each call to iterate().
          Defaults to the length of the data divided by batch_size.
        label: A string that is used to describe this dataset. Usually something
          like 'test' or 'train'.
        '''
        self.label = kwargs.get('label', 'dataset')
        self.number_batches = kwargs.get('batches')
        self.batch = 0

        size = kwargs.get('size', kwargs.get('batch_size', 32))
        batch = None
        cardinality = None
        self.callable = None
        self.batches = None
        if len(data) == 1 and callable(data[0]):
            self.callable = data[0]
            cardinality = '->'
            batch = self.callable()
            if not self.number_batches:
                self.number_batches = size
        else:
            self.batches = [
                [d[i:i + size] for d in data]
                for i in xrange(0, len(data[0]), size)]
            self.shuffle()
            cardinality = len(self.batches)
            batch = self.batches[0]
            if not self.number_batches:
                self.number_batches = cardinality

        logging.info('data %s: %s mini-batches of %s',
            self.label, cardinality, ', '.join(str(x.shape) for x in batch))

    def __iter__(self):
        return self.iterate(True)

    def shuffle(self):
        rng.shuffle(self.batches)

    def iterate(self, update=True):
        if self.callable:
            return self._iter_callable()
        return self._iter_batches(update)

    def _iter_batches(self, update=True):
        k = len(self.batches)
        for b in xrange(self.number_batches):
            yield self.batches[(self.batch + b) % k]
        if update:
            self.update()

    def _iter_callable(self):
        for b in xrange(self.number_batches):
            yield self.callable()

    def update(self):
        if self.callable:
            return
        self.batch += self.number_batches
        if self.batch >= len(self.batches):
            self.shuffle()
            self.batch = 0



# Copyright (c) 2012 Leif Johnson <leif@leifjohnson.net>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

'''This file contains command line flags.'''

import argparse
import climate

g = climate.add_arg_group('Architecture')
g.add_argument('-n', '--layers', nargs='+', type=int, metavar='N',
               help='construct a network with layers of size N1, N2, ...')
g.add_argument('-g', '--activation', default='logistic', metavar='[linear|logistic|tanh|relu]',
               help='function for hidden unit activations')
g.add_argument('-t', '--tied-weights', action='store_true',
               help='tie encoding and decoding weights')
g.add_argument('--decode', type=int, default=1, metavar='N',
               help='decode from the final N layers of the net')

g = climate.add_arg_group('Training')
g.add_argument('-O', '--optimize', default=['sgd'], nargs='+', metavar='[hf|layerwise|sgd|sample]',
               help='train with the given optimization method(s)')
g.add_argument('--no-learn-biases', action='store_true',
               help='if set, do not update bias parameters during learning')
g.add_argument('-u', '--num-updates', type=int, default=128, metavar='N',
               help='perform at most N parameter updates')
g.add_argument('-p', '--patience', type=int, default=15, metavar='N',
               help='stop training if no improvement for N updates')
g.add_argument('-v', '--validate', type=int, default=3, metavar='N',
               help='validate the model every N updates')
g.add_argument('-b', '--batch-size', type=int, default=64, metavar='N',
               help='train with mini-batches of size N')
g.add_argument('-B', '--train-batches', type=int, metavar='N',
               help='use at most N batches during gradient computations')
g.add_argument('-V', '--valid-batches', type=int, metavar='N',
               help='use at most N batches during validation')
g.add_argument('--save-progress', metavar='FILE',
               help='save the model periodically to FILE')

g = climate.add_arg_group('Regularization')
g.add_argument('--input-noise', type=float, default=0, metavar='S',
               help='add noise to network inputs drawn from N(0, S)')
g.add_argument('--input-dropouts', type=float, default=0, metavar='R',
               help='randomly set fraction R of input activations to 0')
g.add_argument('--hidden-noise', type=float, default=0, metavar='S',
               help='add noise to hidden activations drawn from N(0, S)')
g.add_argument('--hidden-dropouts', type=float, default=0, metavar='R',
               help='randomly set fraction R of hidden activations to 0')
g.add_argument('--hidden-l1', type=float, default=0, metavar='K',
               help='regularize hidden activity with K on the L1 term')
g.add_argument('--hidden-l2', type=float, default=0, metavar='K',
               help='regularize hidden activity with K on the L2 term')
g.add_argument('--weight-l1', type=float, default=0, metavar='K',
               help='regularize network weights with K on the L1 term')
g.add_argument('--weight-l2', type=float, default=0, metavar='K',
               help='regularize network weights with K on the L2 term')

g = climate.add_arg_group('SGD Optimization')
g.add_argument('-l', '--learning-rate', type=float, default=0.01, metavar='V',
               help='train the network with a learning rate of V')
g.add_argument('-d', '--learning-rate-decay', type=float, default=0.25, metavar='R',
               help='decay the learning rate by R after stagnant validations')
g.add_argument('-m', '--momentum', type=float, default=0, metavar='V',
               help='train the network with momentum of V')
g.add_argument('--min-improvement', type=float, default=0.01, metavar='R',
               help='train until relative improvement is less than R')
g.add_argument('--gradient-clip', type=float, default=1e5, metavar='R',
               help='clip elementwise gradients to [-R, R]')

g = climate.add_arg_group('HF Optimization')
g.add_argument('-C', '--cg-batches', type=int, metavar='N',
               help='use at most N batches for CG computation')
g.add_argument('--initial-lambda', type=float, default=1., metavar='K',
               help='start the HF method with Tikhonov damping of K')
g.add_argument('--global-backtracking', action='store_true',
               help='backtrack to lowest cost parameters during CG')
g.add_argument('--preconditioner', action='store_true',
               help='precondition the system during CG')

g = climate.add_arg_group('Recurrent Nets')
g.add_argument('--pool-damping', type=float, default=0, metavar='R',
               help='damp recurrent network with R in [0, 1]')
g.add_argument('--pool-noise', type=float, default=0, metavar='S',
               help='add noise to recurrent units drawn from N(0, S)')
g.add_argument('--pool-dropouts', type=float, default=0, metavar='R',
               help='randomly set fraction R of recurrent units to 0')
g.add_argument('--pool-error-start', type=int, default=3, metavar='T',
               help='compute network error starting at time T')



# Copyright (c) 2012 Leif Johnson <leif@leifjohnson.net>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

'''This file contains an object encapsulating a main process.'''

import climate
import theano.tensor as TT

from .dataset import SequenceDataset as Dataset
from . import trainer

logging = climate.get_logger(__name__)


def parse_args(**overrides):
    '''Parse command-line arguments, overriding with keyword arguments.

    Returns an ordered pair of the command-line argument structure, as well as a
    dictionary version of these arguments.
    '''
    args = climate.get_args().parse_args()
    for k, v in overrides.iteritems():
        setattr(args, k, v)
    kwargs = {}
    kwargs.update(vars(args))
    logging.info('runtime arguments:')
    for k in sorted(kwargs):
        logging.info('--%s = %s', k, kwargs[k])
    return args, kwargs


class Experiment(object):
    '''This class encapsulates tasks for training and evaluating a network.
    '''

    TRAINERS = {
        'hf': trainer.HF,
        'sgd': trainer.SGD,
        'sample': trainer.Sample,
        'layerwise': trainer.Layerwise,
        'force': trainer.FORCE,
        }

    def __init__(self, network_class, **overrides):
        '''Set up an experiment -- build a network and a trainer.

        The only input this constructor needs is the Python class of the network
        to build. Other configuration---for example, creating the appropriate
        trainer class---typically takes place by parsing command-line argument
        values.

        Datasets also need to be added to the experiment, either :

        - manually, by calling add_dataset(...), or
        - at runtime, by providing data to the run(train_data, valid_data)
          method.

        Datasets are typically provided as numpy arrays, but they can also be
        provided as callables, as described in the dataset module.

        Any keyword arguments provided to the constructor will be used to
        override values passed on the command line. (Typically this is used to
        provide experiment-specific default values for command line arguments
        that have no global defaults, e.g., network architecture.)
        '''
        self.trainers = []
        self.datasets = {}

        self.args, self.kwargs = parse_args(**overrides)

        kw = {}
        kw.update(self.kwargs)
        self.network = self._build_network(network_class, **kw)

        kw = {}
        kw.update(self.kwargs)
        self._build_trainers(**kw)

    def _build_network(self, network_class, **kwargs):
        '''Build a Network class instance to compute input transformations.
        '''
        activation = self._build_activation()
        if hasattr(activation, '__theanets_name__'):
            logging.info('activation: %s', activation.__theanets_name__)
        del kwargs['activation']
        return network_class(activation=activation, **kwargs)

    def _build_activation(self, act=None):
        '''Given an activation description, return a callable that implements it.
        '''
        def compose(a, b):
            c = lambda z: b(a(z))
            c.__theanets_name__ = '%s(%s)' % (b.__theanets_name__, a.__theanets_name__)
            return c
        act = act or self.args.activation.lower()
        if '+' in act:
            return reduce(compose, (self._build_activation(a) for a in act.split('+')))
        options = {
            'tanh': TT.tanh,
            'linear': lambda z: z,
            'logistic': TT.nnet.sigmoid,
            'softplus': TT.nnet.softplus,

            # shorthands
            'relu': lambda z: TT.maximum(0, z),

            # modifiers
            'rect:max': lambda z: TT.minimum(1, z),
            'rect:min': lambda z: TT.maximum(0, z),

            # normalization
            'norm:dc': lambda z: (z.T - z.mean(axis=1)).T,
            'norm:max': lambda z: (z.T / TT.maximum(1e-10, abs(z).max(axis=1))).T,
            'norm:std': lambda z: (z.T / TT.maximum(1e-10, TT.std(z, axis=1))).T,
            }
        for k, v in options.iteritems():
            v.__theanets_name__ = k
        try:
            return options[act]
        except:
            raise KeyError('unknown --activation %s' % act)

    def _build_trainers(self, **kwargs):
        '''Build trainers from command-line arguments.
        '''
        if not isinstance(self.args.optimize, (tuple, list)):
            self.args.optimize = [self.args.optimize]
        for factory in self.args.optimize:
            self.add_trainer(factory, **kwargs)

    def add_trainer(self, factory, **kwargs):
        '''Add a new trainer to this experiment.

        Arguments:
          factory: The name or Python class of a Trainer.

        Keyword arguments are passed to the trainer factory.
        '''
        if not callable(factory):
            factory = self.TRAINERS[factory]
        kw = {}
        kw.update(self.kwargs)
        kw.update(kwargs)
        self.trainers.append(factory(self.network, **kw))

    def add_dataset(self, label, dataset, **kwargs):
        '''Add a dataset to this experiment.

        The provided label is used to determine the type of data in the set.
        Currently this label can be :

        - train -- for training data,
        - valid -- for validation data, typically a small slice of the training
          data, or
        - cg -- for using the HF optimizer, typically using the same underlying
          data as the training set.

        Other labels can be added, but but they are not currently used.

        The value that you provide for dataset will be encapsulated inside a
        SequenceDataset instance ; see that class for documentation on the types
        of things it needs. In particular, you can currently pass in either a
        list/array/etc. of data, or a callable that generates data dynamically.
        '''
        if 'batches' not in kwargs:
            b = getattr(self.args, '%s_batches' % label, None)
            kwargs['batches'] = b
        if 'size' not in kwargs:
            kwargs['size'] = self.args.batch_size
        kwargs['label'] = label
        if not isinstance(dataset, (tuple, list)):
            dataset = (dataset, )
        self.datasets[label] = Dataset(*dataset, **kwargs)

    def run(self, train=None, valid=None):
        '''Run this experiment by training (and validating) a network.

        Before calling this method, datasets will typically need to have been
        added to the experiment by calling add_dataset(...). However, as a
        shortcut, you can provide training and validation data as arguments to
        this method, and these arguments will be used to add datasets as needed.

        Usually the output of this method is whatever is logged to the console
        during training. After training completes, the network attribute of this
        class will contain the trained network parameters.
        '''
        if train is not None:
            if 'train' not in self.datasets:
                self.add_dataset('train', train)
            if 'cg' not in self.datasets:
                self.add_dataset('cg', train)
        if valid is not None and 'valid' not in self.datasets:
            self.add_dataset('valid', valid)
        for trainer in self.trainers:
            trainer.train(train_set=self.datasets['train'],
                          valid_set=self.datasets['valid'],
                          cg_set=self.datasets['cg'])

    def save(self, path):
        '''Save the parameters in the network to a pickle file on disk.
        '''
        self.network.save(path)

    def load(self, path):
        '''Load the parameters in the network from a pickle file on disk.
        '''
        self.network.load(path)



import os
import setuptools

setuptools.setup(
    name='theanets',
    version='0.1.2',
    packages=setuptools.find_packages(),
    author='Leif Johnson',
    author_email='leif@leifjohnson.net',
    description='A library of neural nets in theano',
    long_description=open(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'README.rst')).read(),
    license='MIT',
    url='http://github.com/lmjohns3/theano-nets/',
    keywords=('machine-learning '
              'neural-network '
              'theano '
              ),
    install_requires=['theano', 'climate'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        ],
    )
